﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;

using RimWorld;
using UnityEngine;
using Verse;
using Verse.Sound;

namespace SlayerWeaponBase
{
    public static class Utility
    {
        public const String TXT_PATH = "debug.txt";
        public static void AddDebug(string text, params object[] args)
        {
#if !DEBUG
                return;
#endif
            
            using (StreamWriter sw = new StreamWriter(TXT_PATH, true, Encoding.Default))
            {
                sw.WriteLine(args.NullOrEmpty() ? text : String.Format(text, args));
            }
        }


        public static void AddDebugEmote(Thing thing, string text, params object[] args)
        {
#if !DEBUG
                return;
#endif
            IntVec3 position = thing.Position;
            float single = (float)thing.Position.y;
            IntVec3 intVec3 = thing.Position;
            MoteMaker.ThrowText(new Vector3((float)position.x + 1f, single, (float)intVec3.z + 1f), thing.Map, args == null ? text : String.Format(text, args), Color.white, -1f);
        }

        public static void ShowEmote(Thing thing, string text, params object[] args)
        {
            IntVec3 position = thing.Position;
            float single = (float)thing.Position.y;
            IntVec3 intVec3 = thing.Position;
            MoteMaker.ThrowText(new Vector3((float)position.x + 1f, single, (float)intVec3.z + 1f), thing.Map, args == null ? text : String.Format(text, args), Color.white, -1f);
        }
    }

    public class ThoughtWorker_IsCarryingRangedWeaponEx : ThoughtWorker_IsCarryingRangedWeapon
    {
        protected override ThoughtState CurrentStateInternal(Pawn p)
        {
            if (p.equipment.Primary != null)
            {
                SlayerWeaponBase wb = p.equipment.Primary.def as SlayerWeaponBase;
                if (!(wb is null) && wb.rg_brawlerLikeThis)
                {
                    return false;
                }
            }
            return base.CurrentStateInternal(p);
        }
    }

    /// <summary>
    /// IEnumerable の拡張メソッドを定義します
    /// </summary>
    public static class IEnumerableExtentionMethods
    {
        /// <summary>
        /// 中身があるかどうか取得します
        /// </summary>
        /// <typeparam name="T">列挙されるタイプ</typeparam>
        /// <param name="enumerable">this</param>
        /// <returns>中身があればTrue、なければFalse</returns>
        public static bool IsEmpty<T>(this IEnumerable<T> enumerable)
        {
            return !enumerable.GetEnumerator().MoveNext();
        }
    }

    public class SecondarySkill
    {
        public SkillDef skill;
        public float ratio;
        public float lernValue;
    }

    public class SlayerSpecialMove
    {
        public string effect;
        public List<string> args;
    }

    public class SlayerProperty
    {
        public string targetDefName;
        public float slayerRatio;
    }

    public class TraitInfo
    {
        public string traitDefName;
        public int degree;
    }

    public class TraitSet
    {
        public TraitDef traitDef;
        public int degree;
    }

    public class SlayerWeaponBase : ThingDef
    {
        private static readonly Dictionary<string, TraitInfo> TraitHash;

        public List<SlayerProperty> ml_slayerProps;
        public int ml_addAttackNum;
        public float ml_skillBonusRatio;
        public List<string> ml_traitAliases;
        public List<TraitInfo> ml_traitInfos;
        public bool ml_colonistOnly;
        public List<SlayerSpecialMove> ml_specialMoves;
        public List<SecondarySkill> ml_secondarySkills;

        public List<SlayerProperty> rg_slayerProps;
        public int rg_addAttackNum;
        public float rg_skillBonusRatio;
        public List<string> rg_traitAliases;
        public List<TraitInfo> rg_traitInfos;
        public bool rg_colonistOnly;
        public bool rg_brawlerLikeThis;
        public List<SlayerSpecialMove> rg_specialMoves;
        public List<SecondarySkill> rg_secondarySkills;

        public static void AddTraitInfoByAlias(List<TraitInfo> traitInfos, params string[] aliases)
        {
            if (traitInfos == null)
            {
                traitInfos = new List<TraitInfo>();
            }
            foreach (string alias in aliases)
            {
                if (TraitHash.ContainsKey(alias))
                {
                    traitInfos.Add(TraitHash[alias]);
                }
            }
        }

        static SlayerWeaponBase()
        {
            TraitHash = new Dictionary<string, TraitInfo>();
            TraitHash["nudist"] = new TraitInfo() { traitDefName = "Nudist", degree = 0 };
            TraitHash["bloodlust"] = new TraitInfo() { traitDefName = "Bloodlust", degree = 0 };
            TraitHash["kind"] = new TraitInfo() { traitDefName = "Kind", degree = 0 };
            TraitHash["psychopath"] = new TraitInfo() { traitDefName = "Psychopath", degree = 0 };
            TraitHash["cannibal"] = new TraitInfo() { traitDefName = "Cannibal", degree = 0 };
            TraitHash["abrasive"] = new TraitInfo() { traitDefName = "Abrasive", degree = 0 };
            TraitHash["tooSmart"] = new TraitInfo() { traitDefName = "TooSmart", degree = 0 };
            TraitHash["brawler"] = new TraitInfo() { traitDefName = "Brawler", degree = 0 };
            TraitHash["masochist"] = new TraitInfo() { traitDefName = "Masochist", degree = 0 };
            TraitHash["prosthophile"] = new TraitInfo() { traitDefName = "Prosthophile", degree = 0 };
            TraitHash["prosthophobe"] = new TraitInfo() { traitDefName = "Prosthophobe", degree = 0 };
            TraitHash["greenThumb"] = new TraitInfo() { traitDefName = "GreenThumb", degree = 0 };
            TraitHash["nightOwl"] = new TraitInfo() { traitDefName = "NightOwl", degree = 0 };
            TraitHash["greedy"] = new TraitInfo() { traitDefName = "Greedy", degree = 0 };
            TraitHash["jealous"] = new TraitInfo() { traitDefName = "Jealous", degree = 0 };
            TraitHash["ascetic"] = new TraitInfo() { traitDefName = "Ascetic", degree = 0 };
            TraitHash["gay"] = new TraitInfo() { traitDefName = "Gay", degree = 0 };
            TraitHash["misandrist"] = new TraitInfo() { traitDefName = "DislikesMen", degree = 0 };
            TraitHash["misogynist"] = new TraitInfo() { traitDefName = "DislikesWomen", degree = 0 };
            TraitHash["annoyingVoice"] = new TraitInfo() { traitDefName = "AnnoyingVoice", degree = 0 };
            TraitHash["creepyBreathing"] = new TraitInfo() { traitDefName = "CreepyBreathing", degree = 0 };
            TraitHash["pyromaniac"] = new TraitInfo() { traitDefName = "Pyromaniac", degree = 0 };
            TraitHash["wimp"] = new TraitInfo() { traitDefName = "Wimp", degree = 0 };
            TraitHash["nimble"] = new TraitInfo() { traitDefName = "Nimble", degree = 0 };
            TraitHash["fastLearner"] = new TraitInfo() { traitDefName = "FastLearner", degree = 0 };
            TraitHash["superImmune"] = new TraitInfo() { traitDefName = "SuperImmune", degree = 0 };
            TraitHash["slowpoke"] = new TraitInfo() { traitDefName = "SpeedOffset", degree = -1 };
            TraitHash["fastWalker"] = new TraitInfo() { traitDefName = "SpeedOffset", degree = 1 };
            TraitHash["jogger"] = new TraitInfo() { traitDefName = "SpeedOffset", degree = 2 };
            TraitHash["chemicalFascination"] = new TraitInfo() { traitDefName = "DrugDesire", degree = 2 };
            TraitHash["chemicalInterest"] = new TraitInfo() { traitDefName = "DrugDesire", degree = 1 };
            TraitHash["teetotaler"] = new TraitInfo() { traitDefName = "DrugDesire", degree = -1 };
            TraitHash["sanguine"] = new TraitInfo() { traitDefName = "NaturalMood", degree = 2 };
            TraitHash["optimist"] = new TraitInfo() { traitDefName = "NaturalMood", degree = 1 };
            TraitHash["pessimist"] = new TraitInfo() { traitDefName = "NaturalMood", degree = -1 };
            TraitHash["depressive"] = new TraitInfo() { traitDefName = "NaturalMood", degree = -2 };
            TraitHash["ironWilled"] = new TraitInfo() { traitDefName = "Nerves", degree = 2 };
            TraitHash["steadfast"] = new TraitInfo() { traitDefName = "Nerves", degree = 1 };
            TraitHash["nervous"] = new TraitInfo() { traitDefName = "Nerves", degree = -1 };
            TraitHash["volatile"] = new TraitInfo() { traitDefName = "Nerves", degree = -2 };
            TraitHash["neurotic"] = new TraitInfo() { traitDefName = "Neurotic", degree = 1 };
            TraitHash["veryNeurotic"] = new TraitInfo() { traitDefName = "Neurotic", degree = 2 };
            TraitHash["industrious"] = new TraitInfo() { traitDefName = "Industriousness", degree = 2 };
            TraitHash["hardWorker"] = new TraitInfo() { traitDefName = "Industriousness", degree = 1 };
            TraitHash["lazy"] = new TraitInfo() { traitDefName = "Industriousness", degree = -1 };
            TraitHash["slothful"] = new TraitInfo() { traitDefName = "Industriousness", degree = -2 };
            TraitHash["psychicallyHypersensitive"] = new TraitInfo() { traitDefName = "PsychicSensitivity", degree = 2 };
            TraitHash["psychicallySensitive"] = new TraitInfo() { traitDefName = "PsychicSensitivity", degree = 1 };
            TraitHash["psychicallyDull"] = new TraitInfo() { traitDefName = "PsychicSensitivity", degree = -1 };
            TraitHash["psychicallyDeaf"] = new TraitInfo() { traitDefName = "PsychicSensitivity", degree = -2 };
            TraitHash["carefulShooter"] = new TraitInfo() { traitDefName = "ShootingAccuracy", degree = 1 };
            TraitHash["triggerHappy"] = new TraitInfo() { traitDefName = "ShootingAccuracy", degree = -1 };
            TraitHash["beautiful"] = new TraitInfo() { traitDefName = "Beauty", degree = 2 };
            TraitHash["pretty"] = new TraitInfo() { traitDefName = "Beauty", degree = 1 };
            TraitHash["ugly"] = new TraitInfo() { traitDefName = "Beauty", degree = -1 };
            TraitHash["staggeringlyUgly"] = new TraitInfo() { traitDefName = "Beauty", degree = -2 };
        }
    }

    public abstract class SWB_Verb_MeleeAttackBase : Verb_MeleeAttack
    {
        public const string PROP_BASE_MLW_COOLDOWN = "MeleeWeapon_Cooldown";
        public const string MLW_COOLDOWN_F = "meleeWeaponCooldownFast";
        public const string ADD_ATTACK_NUM = "addAttackNum";
        public const string RACE_HUMANLIKE = "Humanlike";
        public const string RACE_ANIMAL = "Animal";
        public const string RACE_MECHANOID = "Mechanoid";
        public const string RACE_INSECTOID = "Insectoid";

        #region フィールド
        public int m_skillBonusRatio = 10;
        public bool m_ColonistOnly = false;

        public static Dictionary<string, float> m_CooldownDefaults = new Dictionary<string, float>();
        #endregion

        #region プロパティ
        public LocalTargetInfo CurrentTarget
        {
            get
            {
                return this.currentTarget;
            }
        }

        private Dictionary<Pawn, int> _AttackCountMax = new Dictionary<Pawn, int>();
        public int m_AttackCountMax
        {
            get
            {
                if (this.CasterPawn == null)
                {
                    return 0;
                }
                if (!this._AttackCountMax.ContainsKey(this.CasterPawn))
                {
                    this._AttackCountMax[this.CasterPawn] = 0;
                }
                return this._AttackCountMax[this.CasterPawn];
            }
            set
            {
                this._AttackCountMax[this.CasterPawn] = value;
            }
        }

        private Dictionary<Pawn, int> _AttackCount = new Dictionary<Pawn, int>();
        public int m_AttackCount
        {
            get
            {
                if (this.CasterPawn == null)
                {
                    return 0;
                }
                if (!this._AttackCount.ContainsKey(this.CasterPawn))
                {
                    this._AttackCount[this.CasterPawn] = 0;
                }
                return this._AttackCount[this.CasterPawn];
            }
            set
            {
                this._AttackCount[this.CasterPawn] = value;
            }
        }

        private Dictionary<Pawn, int> _AddAttackNum = new Dictionary<Pawn, int>();
        public int m_AddAttackNum
        {
            get
            {
                if (this.CasterPawn == null)
                {
                    return 0;
                }
                if (!this._AddAttackNum.ContainsKey(this.CasterPawn))
                {
                    this._AddAttackNum[this.CasterPawn] = 0;
                }
                return this._AddAttackNum[this.CasterPawn];
            }
            set
            {
                this._AddAttackNum[this.CasterPawn] = value;
            }
        }

        private Dictionary<string, bool> _InitConfigured = new Dictionary<string, bool>();
        public bool m_InitConfigured
        {
            get
            {
                if (this.ownerEquipment == null)
                {
                    return false;
                }
                string defName = this.ownerEquipment.def.defName;
                if (!this._InitConfigured.ContainsKey(defName))
                {
                    this._InitConfigured[defName] = false;
                }
                return this._InitConfigured[defName];
            }
            set
            {
                if (this.ownerEquipment == null)
                {
                    return;
                }
                string defName = this.ownerEquipment.def.defName;
                this._InitConfigured[defName] = value;
            }
        }

        private Dictionary<string, Dictionary<string, float>> _Slayers = new Dictionary<string, Dictionary<string, float>>();
        public Dictionary<string, float> m_Slayers
        {
            get
            {
                if (this.ownerEquipment == null)
                {
                    return null;
                }
                string defName = this.ownerEquipment.def.defName;
                if (!this._Slayers.ContainsKey(defName))
                {
                    this._Slayers[defName] = new Dictionary<string, float>();
                }
                return this._Slayers[defName];
            }
            set
            {
                if (this.ownerEquipment == null)
                {
                    return;
                }
                string defName = this.ownerEquipment.def.defName;
                this._Slayers[defName] = value;
            }
        }

        private Dictionary<string, List<TraitSet>> _Traits = new Dictionary<string, List<TraitSet>>();
        public List<TraitSet> m_Traits
        {
            get
            {
                if (this.ownerEquipment == null)
                {
                    return null;
                }
                string defName = this.ownerEquipment.def.defName;
                if (!this._Traits.ContainsKey(defName))
                {
                    this._Traits[defName] = new List<TraitSet>();
                }
                return this._Traits[defName];
            }
            set
            {
                if (this.ownerEquipment == null)
                {
                    return;
                }
                string defName = this.ownerEquipment.def.defName;
                this._Traits[defName] = value;
            }
        }

        private Dictionary<string, List<SecondarySkill>> _SecondarySkills = new Dictionary<string, List<SecondarySkill>>();
        public List<SecondarySkill> m_SecondarySkills
        {
            get
            {
                if (this.ownerEquipment == null)
                {
                    return null;
                }
                string defName = this.ownerEquipment.def.defName;
                if (!this._SecondarySkills.ContainsKey(defName))
                {
                    this._SecondarySkills[defName] = new List<SecondarySkill>();
                }
                return this._SecondarySkills[defName];
            }
            set
            {
                if (this.ownerEquipment == null)
                {
                    return;
                }
                string defName = this.ownerEquipment.def.defName;
                this._SecondarySkills[defName] = value;
            }
        }


        #endregion


        #region メソッド

        protected void ResetCount()
        {
            m_AttackCount = 0;
        }

        protected void SetCooldown(SlayerWeaponBase uniqueProperties, float coolDown)
        {
            foreach (ThingWithComps th in base.CasterPawn.equipment.AllEquipmentListForReading)
            {
                if (th.def.defName == uniqueProperties.defName)
                {
                    float max = GetCooldownDefault();
                    foreach(Tool tool in th.def.tools)
                    {
                        max = Math.Max(max, tool.cooldownTime);
                        tool.cooldownTime = coolDown;
                    }
                    SetCooldownDefault(max);
                }
            }
        }


        protected void SetCooldownDefault(float defaultCooldown)
        {
            ThingDef equipWeapon = this.ownerEquipment.def;
            if (!m_CooldownDefaults.ContainsKey(equipWeapon.defName))
            {
                m_CooldownDefaults[equipWeapon.defName] = defaultCooldown;
                //AddDebug(String.Format("{0}のデフォルトクールダウンは{1}ですよ。", equipWeapon.defName, defaultCooldown));
            }
        }

        protected float GetCooldownDefault()
        {
            float ret = 2f;
            ThingDef equipWeapon = this.ownerEquipment.def;
            if (m_CooldownDefaults.ContainsKey(equipWeapon.defName))
            {
                ret = m_CooldownDefaults[equipWeapon.defName];
            }
            return ret;
        }

        protected bool HasTrait(Trait trait, int degree = 0)
        {
            if (trait is null)
                return false;
            return trait.Degree == degree;
        }

        protected virtual float GetNonMissChance(LocalTargetInfo target)
        {
            if (this.surpriseAttack)
            {
                return 1f;
            }
            if (this.IsTargetImmobile(target))
            {
                return 1f;
            }
            return base.CasterPawn.GetStatValue(StatDefOf.MeleeHitChance, true);
        }

        protected virtual float GetDodgeChance(LocalTargetInfo target)
        {
            if (this.surpriseAttack)
            {
                return 0f;
            }
            if (this.IsTargetImmobile(target))
            {
                return 0f;
            }
            Pawn pawn = target.Thing as Pawn;
            if (pawn == null)
            {
                return 0f;
            }
            Stance_Busy stance_Busy = pawn.stances.curStance as Stance_Busy;
            if (stance_Busy != null && stance_Busy.verb != null && !stance_Busy.verb.verbProps.MeleeRange)
            {
                return 0f;
            }
            return pawn.GetStatValue(StatDefOf.MeleeDodgeChance, true);
        }

        protected virtual bool IsTargetImmobile(LocalTargetInfo target)
        {
            Thing thing = target.Thing;
            Pawn pawn = thing as Pawn;
            return thing.def.category != ThingCategory.Pawn || pawn.Downed || pawn.GetPosture() != PawnPosture.Standing;
        }

        protected virtual SoundDef SoundHitPawn()
        {
            if (this.ownerEquipment != null && this.ownerEquipment.Stuff != null)
            {
                if (this.verbProps.meleeDamageDef.armorCategory == DamageArmorCategoryDefOf.Sharp)
                {
                    if (!this.ownerEquipment.Stuff.stuffProps.soundMeleeHitSharp.NullOrUndefined())
                    {
                        return this.ownerEquipment.Stuff.stuffProps.soundMeleeHitSharp;
                    }
                }
                else if (!this.ownerEquipment.Stuff.stuffProps.soundMeleeHitBlunt.NullOrUndefined())
                {
                    return this.ownerEquipment.Stuff.stuffProps.soundMeleeHitBlunt;
                }
            }
            if (base.CasterPawn != null && !base.CasterPawn.def.race.soundMeleeHitPawn.NullOrUndefined())
            {
                return base.CasterPawn.def.race.soundMeleeHitPawn;
            }
            return SoundDefOf.Pawn_Melee_Punch_HitPawn;
        }

        protected virtual SoundDef SoundHitBuilding()
        {
            if (this.ownerEquipment != null && this.ownerEquipment.Stuff != null)
            {
                if (this.verbProps.meleeDamageDef.armorCategory == DamageArmorCategoryDefOf.Sharp)
                {
                    if (!this.ownerEquipment.Stuff.stuffProps.soundMeleeHitSharp.NullOrUndefined())
                    {
                        return this.ownerEquipment.Stuff.stuffProps.soundMeleeHitSharp;
                    }
                }
                else if (!this.ownerEquipment.Stuff.stuffProps.soundMeleeHitBlunt.NullOrUndefined())
                {
                    return this.ownerEquipment.Stuff.stuffProps.soundMeleeHitBlunt;
                }
            }
            if (base.CasterPawn != null && !base.CasterPawn.def.race.soundMeleeHitBuilding.NullOrUndefined())
            {
                return base.CasterPawn.def.race.soundMeleeHitBuilding;
            }
            return SoundDefOf.Pawn_Melee_Punch_HitBuilding;
        }

        protected virtual SoundDef SoundMiss()
        {
            if (base.CasterPawn != null && !base.CasterPawn.def.race.soundMeleeMiss.NullOrUndefined())
            {
                return base.CasterPawn.def.race.soundMeleeMiss;
            }
            return SoundDefOf.Pawn_Melee_Punch_Miss;
        }
        #endregion
    }

    public class CasterPos
    {
        public int x;
        public int y;
        public int z;
    }

    public abstract class SWB_Verb_RangedAttackBase : Verb_Shoot
    {
        public const string PROP_BASE_RGW_COOLDOWN = "RangedWeapon_Cooldown";
        public const string ADD_ATTACK_NUM = "addAttackNum";
        public const string RACE_HUMANLIKE = "Humanlike";
        public const string RACE_ANIMAL = "Animal";
        public const string RACE_MECHANOID = "Mechanoid";
        public const string RACE_INSECTOID = "Insectoid";

        #region フィールド
        public int m_skillBonusRatio = 10;
        public bool m_ColonistOnly = false;
        public int m_BurstShotCountDefault;

        public static Dictionary<string, float> m_CooldownDefaults = new Dictionary<string, float>();
        #endregion

        #region プロパティ
        public LocalTargetInfo CurrentTarget
        {
            get
            {
                return this.currentTarget;
            }
        }

        private Dictionary<Pawn, int> _AttackCountMax = new Dictionary<Pawn, int>();
        public int m_AttackCountMax
        {
            get
            {
                if (this.CasterPawn == null)
                {
                    return 0;
                }
                if (!this._AttackCountMax.ContainsKey(this.CasterPawn))
                {
                    this._AttackCountMax[this.CasterPawn] = 0;
                }
                return this._AttackCountMax[this.CasterPawn];
            }
            set
            {
                this._AttackCountMax[this.CasterPawn] = value;
            }
        }

        private Dictionary<Pawn, int> _AttackCount = new Dictionary<Pawn, int>();
        public int m_AttackCount
        {
            get
            {
                if (this.CasterPawn == null)
                {
                    return 0;
                }
                if (!this._AttackCount.ContainsKey(this.CasterPawn))
                {
                    this._AttackCount[this.CasterPawn] = 0;
                }
                return this._AttackCount[this.CasterPawn];
            }
            set
            {
                this._AttackCount[this.CasterPawn] = value;
            }
        }

        private Dictionary<Pawn, int> _AddAttackNum = new Dictionary<Pawn, int>();
        public int m_AddAttackNum
        {
            get
            {
                if (this.CasterPawn == null)
                {
                    return 0;
                }
                if (!this._AddAttackNum.ContainsKey(this.CasterPawn))
                {
                    this._AddAttackNum[this.CasterPawn] = 0;
                }
                return this._AddAttackNum[this.CasterPawn];
            }
            set
            {
                this._AddAttackNum[this.CasterPawn] = value;
            }
        }

        private Dictionary<string, bool> _InitConfigured = new Dictionary<string, bool>();
        public bool m_InitConfigured
        {
            get
            {
                if (this.ownerEquipment == null)
                {
                    return false;
                }
                string defName = this.ownerEquipment.def.defName;
                if (!this._InitConfigured.ContainsKey(defName))
                {
                    this._InitConfigured[defName] = false;
                }
                return this._InitConfigured[defName];
            }
            set
            {
                if (this.ownerEquipment == null)
                {
                    return;
                }
                string defName = this.ownerEquipment.def.defName;
                this._InitConfigured[defName] = value;
            }
        }

        private Dictionary<string, Dictionary<string, float>> _Slayers = new Dictionary<string, Dictionary<string, float>>();
        public Dictionary<string, float> m_Slayers
        {
            get
            {
                if (this.ownerEquipment == null)
                {
                    return null;
                }
                string defName = this.ownerEquipment.def.defName;
                if (!this._Slayers.ContainsKey(defName))
                {
                    this._Slayers[defName] = new Dictionary<string, float>();
                }
                return this._Slayers[defName];
            }
            set
            {
                if (this.ownerEquipment == null)
                {
                    return;
                }
                string defName = this.ownerEquipment.def.defName;
                this._Slayers[defName] = value;
            }
        }

        private Dictionary<string, List<TraitSet>> _Traits = new Dictionary<string, List<TraitSet>>();
        public List<TraitSet> m_Traits
        {
            get
            {
                if (this.ownerEquipment == null)
                {
                    return null;
                }
                string defName = this.ownerEquipment.def.defName;
                if (!this._Traits.ContainsKey(defName))
                {
                    this._Traits[defName] = new List<TraitSet>();
                }
                return this._Traits[defName];
            }
            set
            {
                if (this.ownerEquipment == null)
                {
                    return;
                }
                string defName = this.ownerEquipment.def.defName;
                this._Traits[defName] = value;
            }
        }

        private Dictionary<string, List<SecondarySkill>> _SecondarySkills = new Dictionary<string, List<SecondarySkill>>();
        public List<SecondarySkill> m_SecondarySkills
        {
            get
            {
                if (this.ownerEquipment == null)
                {
                    return null;
                }
                string defName = this.ownerEquipment.def.defName;
                if (!this._SecondarySkills.ContainsKey(defName))
                {
                    this._SecondarySkills[defName] = new List<SecondarySkill>();
                }
                return this._SecondarySkills[defName];
            }
            set
            {
                if (this.ownerEquipment == null)
                {
                    return;
                }
                string defName = this.ownerEquipment.def.defName;
                this._SecondarySkills[defName] = value;
            }
        }

        protected Dictionary<Pawn, CasterPos> _PrevPos = new Dictionary<Pawn, CasterPos>();
        public CasterPos m_PrevPos
        {
            get
            {
                if (this.CasterPawn == null)
                {
                    return null;
                }
                if (!this._PrevPos.ContainsKey(this.CasterPawn))
                {
                    this._PrevPos[this.CasterPawn] = new CasterPos() { x = 0, y = 0, z = 0 };
                }
                return this._PrevPos[this.CasterPawn];
            }
            set
            {
                this._PrevPos[this.CasterPawn] = value;
            }
        }

        #endregion

        #region Clone Methods
        protected ThingDef GetClonedProjectileDef(ThingDef org)
        {
            ThingDef ret = new ThingDef();
            ret.affectsRegions = org.affectsRegions;
            ret.altitudeLayer = org.altitudeLayer;
            ret.alwaysHaulable = org.alwaysHaulable;
            ret.apparel = org.apparel;
            ret.blockLight = org.blockLight;
            ret.blockPlants = org.blockPlants;
            ret.blockWind = org.blockWind;
            ret.blueprintClass = org.blueprintClass;
            ret.blueprintDef = org.blueprintDef;
            ret.blueprintGraphicData = org.blueprintGraphicData;
            ret.building = org.building;
            ret.buildingPrerequisites = org.buildingPrerequisites;
            ret.butcherProducts = org.butcherProducts;
            ret.canBeSpawningInventory = org.canBeSpawningInventory;
            ret.castEdgeShadows = org.castEdgeShadows;
            ret.category = org.category;
            ret.colorGenerator = org.colorGenerator;
            ret.colorGeneratorInTraderStock = org.colorGeneratorInTraderStock;
            ret.comps = org.comps;
            ret.constructEffect = org.constructEffect;
            ret.costList = org.costList;
            ret.costStuffCount = org.costStuffCount;
            ret.coversFloor = org.coversFloor;
            ret.damageMultipliers = org.damageMultipliers;
            ret.debugRandomId = org.debugRandomId;
            ret.deepCommonality = org.deepCommonality;
            ret.deepCountPerCell = org.deepCountPerCell;
            ret.defaultPlacingRot = org.defaultPlacingRot;
            ret.defName = org.defName;
            ret.description = org.description;
            ret.designateHaulable = org.designateHaulable;
            ret.designationCategory = org.designationCategory;
            ret.designationHotKey = org.designationHotKey;
            ret.destroyable = org.destroyable;
            ret.destroyOnDrop = org.destroyOnDrop;
            ret.drawDamagedOverlay = org.drawDamagedOverlay;
            ret.drawerType = org.drawerType;
            ret.drawGUIOverlay = org.drawGUIOverlay;
            ret.drawOffscreen = org.drawOffscreen;
            ret.drawPlaceWorkersWhileSelected = org.drawPlaceWorkersWhileSelected;
            ret.entityDefToBuild = org.entityDefToBuild;
            ret.equipmentType = org.equipmentType;
            ret.equippedAngleOffset = org.equippedAngleOffset;
            ret.equippedStatOffsets = org.equippedStatOffsets;
            ret.fertility = org.fertility;
            ret.fillPercent = org.fillPercent;
            ret.filth = org.filth;
            ret.filthLeaving = org.filthLeaving;
            ret.forceDebugSpawnable = org.forceDebugSpawnable;
            ret.frameDef = org.frameDef;
            ret.gas = org.gas;
            ret.generateAllowChance = org.generateAllowChance;
            ret.generateCommonality = org.generateCommonality;
            ret.graphic = org.graphic;
            ret.graphicData = org.graphicData;
            ret.hasInteractionCell = org.hasInteractionCell;
            ret.hasTooltip = org.hasTooltip;
            ret.hideAtSnowDepth = org.hideAtSnowDepth;
            ret.holdsRoof = org.holdsRoof;
            ret.iconDrawScale = org.iconDrawScale;
            ret.index = org.index;
            ret.ingestible = org.ingestible;
            ret.inspectorTabs = org.inspectorTabs;
            ret.inspectorTabsResolved = org.inspectorTabsResolved;
            ret.installBlueprintDef = org.installBlueprintDef;
            ret.interactionCellOffset = org.interactionCellOffset;
            ret.intricate = org.intricate;
            ret.isAutoAttackableMapObject = org.isAutoAttackableMapObject;
            ret.isBodyPartOrImplant = org.isBodyPartOrImplant;
            ret.isFrame = org.isFrame;
            ret.isSaveable = org.isSaveable;
            ret.isUnfinishedThing = org.isUnfinishedThing;
            ret.killedLeavings = org.killedLeavings;
            ret.label = org.label;
            ret.leaveResourcesWhenKilled = org.leaveResourcesWhenKilled;
            ret.leaveTerrain = org.leaveTerrain;
            ret.menuHidden = org.menuHidden;
            ret.mineable = org.mineable;
            ret.minifiedDef = org.minifiedDef;
            ret.modExtensions = org.modExtensions;
            ret.mote = org.mote;
            ret.naturalTerrain = org.naturalTerrain;
            ret.neverMultiSelect = org.neverMultiSelect;
            ret.neverOverlapFloors = org.neverOverlapFloors;
            ret.passability = org.passability;
            ret.pathCost = org.pathCost;
            ret.pathCostIgnoreRepeat = org.pathCostIgnoreRepeat;
            ret.placeWorkers = org.placeWorkers;
            ret.placingDraggableDimensions = org.placingDraggableDimensions;
            ret.plant = org.plant;
            //ret.projectile = org.projectile;

            ProjectileProperties pp = new ProjectileProperties();
            pp.alwaysFreeIntercept = org.projectile.alwaysFreeIntercept;
            pp.damageAmountBase = org.projectile.damageAmountBase;
            pp.damageDef = org.projectile.damageDef;
            pp.explosionDelay = org.projectile.explosionDelay;
            pp.explosionRadius = org.projectile.explosionRadius;
            pp.flyOverhead = org.projectile.flyOverhead;
            pp.postExplosionSpawnThingDef = org.projectile.postExplosionSpawnThingDef;
            pp.preExplosionSpawnThingDef = org.projectile.preExplosionSpawnThingDef;
            pp.soundAmbient = org.projectile.soundAmbient;
            pp.soundExplode = org.projectile.soundExplode;
            pp.soundHitThickRoof = org.projectile.soundHitThickRoof;
            pp.soundImpactAnticipate = org.projectile.soundImpactAnticipate;
            pp.speed = org.projectile.speed;

            pp.ai_IsIncendiary = org.projectile.ai_IsIncendiary;
            pp.applyDamageToExplosionCellsNeighbors = org.projectile.applyDamageToExplosionCellsNeighbors;
            pp.explosionChanceToStartFire = org.projectile.explosionChanceToStartFire;
            pp.explosionDealMoreDamageAtCenter = org.projectile.explosionDealMoreDamageAtCenter;
            pp.explosionEffect = org.projectile.explosionEffect;
            pp.postExplosionSpawnChance = org.projectile.postExplosionSpawnChance;
            pp.postExplosionSpawnThingCount = org.projectile.postExplosionSpawnThingCount;
            pp.preExplosionSpawnChance = org.projectile.preExplosionSpawnChance;
            pp.preExplosionSpawnThingCount = org.projectile.preExplosionSpawnThingCount;

            ret.projectile = pp;

            ret.race = org.race;
            ret.randomizeRotationOnSpawn = org.randomizeRotationOnSpawn;
            ret.recipeMaker = org.recipeMaker;
            ret.recipes = org.recipes;
            ret.repairEffect = org.repairEffect;
            ret.researchPrerequisites = org.researchPrerequisites;
            ret.resourceReadoutAlwaysShow = org.resourceReadoutAlwaysShow;
            ret.resourceReadoutPriority = org.resourceReadoutPriority;
            ret.resourcesFractionWhenDeconstructed = org.resourcesFractionWhenDeconstructed;
            ret.rotatable = org.rotatable;
            ret.saveCompressible = org.saveCompressible;
            ret.scatterableOnMapGen = org.scatterableOnMapGen;
            ret.seeThroughFog = org.seeThroughFog;
            ret.selectable = org.selectable;
            ret.shortHash = org.shortHash;
            ret.size = org.size;
            ret.slagDef = org.slagDef;
            ret.smallVolume = org.smallVolume;
            ret.smeltable = org.smeltable;
            ret.smeltProducts = org.smeltProducts;
            ret.socialPropernessMatters = org.socialPropernessMatters;
            ret.soundDrop = org.soundDrop;
            ret.soundImpactDefault = org.soundImpactDefault;
            ret.soundInteract = org.soundInteract;
            ret.soundPickup = org.soundPickup;
            ret.specialDisplayRadius = org.specialDisplayRadius;
            ret.stackLimit = org.stackLimit;
            ret.startingHpRange = org.startingHpRange;
            ret.statBases = org.statBases;
            ret.staticSunShadowHeight = org.staticSunShadowHeight;
            ret.stealable = org.stealable;
            ret.storedConceptLearnOpportunity = org.storedConceptLearnOpportunity;
            ret.stuffCategories = org.stuffCategories;
            ret.stuffProps = org.stuffProps;
            ret.surfaceType = org.surfaceType;
            ret.techHediffsTags = org.techHediffsTags;
            ret.techLevel = org.techLevel;
            ret.terrainAffordanceNeeded = org.terrainAffordanceNeeded;
            ret.thingCategories = org.thingCategories;
            ret.thingClass = org.thingClass;
            ret.tickerType = org.tickerType;
            ret.tradeability = org.tradeability;
            ret.tradeNeverStack = org.tradeNeverStack;
            ret.tradeTags = org.tradeTags;
            ret.uiIcon = org.uiIcon;
            ret.uiIconPath = org.uiIconPath;
            ret.useHitPoints = org.useHitPoints;
            ret.weaponTags = org.weaponTags;
            return ret;
        }
        #endregion

        protected bool IsTargetImmobile(LocalTargetInfo target)
        {
            Thing thing = target.Thing;
            Pawn pawn = thing as Pawn;
            return thing.def.category != ThingCategory.Pawn || pawn.Downed || pawn.GetPosture() != PawnPosture.Standing;
        }

        protected void SetCooldownDefault(float defaultCooldown)
        {
            ThingDef equipWeapon = this.ownerEquipment.def;
            if (!m_CooldownDefaults.ContainsKey(equipWeapon.defName))
            {
                m_CooldownDefaults[equipWeapon.defName] = defaultCooldown;
                //AddDebug(String.Format("{0}のデフォルトクールダウンは{1}ですよ。", equipWeapon.defName, defaultCooldown));
            }
        }

        protected float GetCooldownDefault()
        {
            float ret = 2f;
            ThingDef equipWeapon = this.ownerEquipment.def;
            if (m_CooldownDefaults.ContainsKey(equipWeapon.defName))
            {
                ret = m_CooldownDefaults[equipWeapon.defName];
            }
            return ret;
        }

        protected void ResetCount()
        {
            m_AttackCount = 0;
        }

        protected void SetCooldown(SlayerWeaponBase uniqueProperties, float coolDown)
        {
            foreach (ThingWithComps th in base.CasterPawn.equipment.AllEquipmentListForReading)
            {
                if (th.def.defName == uniqueProperties.defName)
                {
                    float max = GetCooldownDefault();
                    foreach (Tool tool in th.def.tools)
                    {
                        max = Math.Max(max, tool.cooldownTime);
                        tool.cooldownTime = coolDown;
                    }
                    SetCooldownDefault(max);
                }
            }
        }

        protected bool HasTrait(Trait trait, int degree = 0)
        {
            if (trait is null)
                return false;
            return trait.Degree == degree;
        }


        protected override bool TryCastShot()
        {
            bool flag = TryCastShotImpl();
            if (flag && base.CasterIsPawn)
            {
                base.CasterPawn.records.Increment(RecordDefOf.ShotsFired);
            }
            return flag;
        }

        protected virtual bool TryCastShotImpl()
        {
            if (this.currentTarget.HasThing && this.currentTarget.Thing.Map != this.caster.Map)
            {
                return false;
            }
            ThingDef projectile = this.Projectile;
            if (projectile == null)
            {
                return false;
            }
            ShootLine shootLine;
            bool flag = base.TryFindShootLineFromTo(this.caster.Position, this.currentTarget, out shootLine);
            if (this.verbProps.stopBurstWithoutLos && !flag)
            {
                return false;
            }
            if (this.ownerEquipment != null)
            {
                CompChangeableProjectile comp = this.ownerEquipment.GetComp<CompChangeableProjectile>();
                if (comp != null)
                {
                    comp.Notify_ProjectileLaunched();
                }
            }
            Thing launcher = this.caster;
            Thing equipment = this.ownerEquipment;
            CompMannable compMannable = this.caster.TryGetComp<CompMannable>();
            if (compMannable != null && compMannable.ManningPawn != null)
            {
                launcher = compMannable.ManningPawn;
                equipment = this.caster;
            }
            Vector3 drawPos = this.caster.DrawPos;
            Projectile projectile2 = (Projectile)GenSpawn.Spawn(projectile, shootLine.Source, this.caster.Map);
            projectile2.FreeIntercept = (this.canFreeInterceptNow && !projectile2.def.projectile.flyOverhead);
            if (this.verbProps.forcedMissRadius > 0.5f)
            {
                float num = (float)(this.currentTarget.Cell - this.caster.Position).LengthHorizontalSquared;
                float num2;
                if (num < 9f)
                {
                    num2 = 0f;
                }
                else if (num < 25f)
                {
                    num2 = this.verbProps.forcedMissRadius * 0.5f;
                }
                else if (num < 49f)
                {
                    num2 = this.verbProps.forcedMissRadius * 0.8f;
                }
                else
                {
                    num2 = this.verbProps.forcedMissRadius * 1f;
                }
                if (num2 > 0.5f)
                {
                    int max = GenRadial.NumCellsInRadius(this.verbProps.forcedMissRadius);
                    int num3 = Rand.Range(0, max);
                    if (num3 > 0)
                    {
                        if (DebugViewSettings.drawShooting)
                        {
                            MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToForRad", -1f);
                        }
                        IntVec3 c = this.currentTarget.Cell + GenRadial.RadialPattern[num3];
                        if (this.currentTarget.HasThing)
                        {
                            projectile2.ThingToNeverIntercept = this.currentTarget.Thing;
                        }
                        if (!projectile2.def.projectile.flyOverhead)
                        {
                            projectile2.InterceptWalls = true;
                        }
                        projectile2.Launch(launcher, drawPos, c, equipment, this.currentTarget.Thing);
                        return true;
                    }
                }
            }
            ShotReport shotReport = ShotReport.HitReportFor(this.caster, this, this.currentTarget);
            if (Rand.Value > shotReport.ChanceToNotGoWild_IgnoringPosture)
            {
                if (DebugViewSettings.drawShooting)
                {
                    MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToWild", -1f);
                }
                shootLine.ChangeDestToMissWild();
                if (this.currentTarget.HasThing)
                {
                    projectile2.ThingToNeverIntercept = this.currentTarget.Thing;
                }
                if (!projectile2.def.projectile.flyOverhead)
                {
                    projectile2.InterceptWalls = true;
                }
                projectile2.Launch(launcher, drawPos, shootLine.Dest, equipment, this.currentTarget.Thing);
                return true;
            }
            if (Rand.Value > shotReport.ChanceToNotHitCover)
            {
                if (DebugViewSettings.drawShooting)
                {
                    MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToCover", -1f);
                }
                if (this.currentTarget.Thing != null && this.currentTarget.Thing.def.category == ThingCategory.Pawn)
                {
                    Thing randomCoverToMissInto = shotReport.GetRandomCoverToMissInto();
                    if (!projectile2.def.projectile.flyOverhead)
                    {
                        projectile2.InterceptWalls = true;
                    }
                    projectile2.Launch(launcher, drawPos, randomCoverToMissInto, equipment, this.currentTarget.Thing);
                    return true;
                }
            }
            if (DebugViewSettings.drawShooting)
            {
                MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToHit", -1f);
            }
            if (!projectile2.def.projectile.flyOverhead)
            {
                projectile2.InterceptWalls = (!this.currentTarget.HasThing || this.currentTarget.Thing.def.Fillage == FillCategory.Full);
            }
            if (this.currentTarget.Thing != null)
            {
                projectile2.Launch(launcher, drawPos, this.currentTarget, equipment, this.currentTarget.Thing);
            }
            else
            {
                projectile2.Launch(launcher, drawPos, shootLine.Dest, equipment, this.currentTarget.Thing);
            }
            return true;
        }
    }

}
