﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using Harmony;
using RimWorld;
using UnityEngine;
using Verse;
using Verse.Sound;

namespace SlayerWeaponBase
{


    public enum SpecialMoveLocationML
    {
        NONE = 0,
        BEFORE_EVAL = 1,
        BEFORE_DAM = 101,
        BEFORE_END = 201,
    }

    public enum SpecialMoveLocationRG
    {
        NONE = 1000,
        BEFORE_EVAL = 1001,
        BEFORE_DAM = 1101,
        BEFORE_END = 1201,
    }

    public class SpecialMoveController
    {
        public const string ALIAS_SPM_DELICA43 = "DELICA43";
        public const string ALIAS_SPM_RELOAD_ON_ENEMY_DOWN = "RELOAD_ON_ENEMY_DOWN";
        public const string ALIAS_SPM_RELOAD_ON_ENEMY_DOWN_ML = "RELOAD_ON_ENEMY_DOWN_ML";
        public const string ALIAS_SPM_RELOAD_ON_ENEMY_DOWN_RG = "RELOAD_ON_ENEMY_DOWN_RG";

        private static Dictionary<string, SpecialMoveBase> _InstanceHash;

        static SpecialMoveController()
        {
            InitInstanceHash();
        }

        private static void InitInstanceHash()
        {
            _InstanceHash = new Dictionary<string, SpecialMoveBase>();
            _InstanceHash[ALIAS_SPM_DELICA43] = new SPM_Delica43();
            _InstanceHash[ALIAS_SPM_RELOAD_ON_ENEMY_DOWN] = new SPM_ReloadOnEnemeyDown();
            _InstanceHash[ALIAS_SPM_RELOAD_ON_ENEMY_DOWN_ML] = new SPM_ReloadOnEnemeyDownML();
            _InstanceHash[ALIAS_SPM_RELOAD_ON_ENEMY_DOWN_RG] = new SPM_ReloadOnEnemeyDownRG();
        }
        public static void AddInstanceHash<T>(string hashKey)
            where T : SpecialMoveBase, new()
        {
            _InstanceHash[hashKey] = new T();
        }

        public static SpecialMoveBase GetSpecialMoveInstance(string hashKey)
        {
            if (_InstanceHash.ContainsKey(hashKey))
            {
                return _InstanceHash[hashKey];
            }
            return null;
        }

        public static void DoSpecialMove(SWB_Verb_MeleeAttack melee, SpecialMoveLocationML loc, List<SlayerSpecialMove> sspms)
        {
            if (sspms == null)
                return;
            foreach(SlayerSpecialMove sspm in sspms)
            {
                DoSpecialMove(melee, loc, sspm);
            }
        }

        public static void DoSpecialMove(SWB_Verb_MeleeAttack melee, SpecialMoveLocationML loc, SlayerSpecialMove sspm)
        {
            if (sspm == null)
                return;
            SpecialMoveBase effectInstance = GetSpecialMoveInstance(sspm.effect);
            if (effectInstance == null)
                return;
            if (effectInstance.GetLocationML == loc)
            {
                effectInstance.DoSpecialMove(melee, sspm.args);
            }
        }

        public static void DoSpecialMove(SWB_Verb_RangedAttack range, SpecialMoveLocationRG loc, List<SlayerSpecialMove> sspms)
        {
            if (sspms == null)
                return;
            foreach (SlayerSpecialMove sspm in sspms)
            {
                DoSpecialMove(range, loc, sspm);
            }
        }

        public static void DoSpecialMove(SWB_Verb_RangedAttack range, SpecialMoveLocationRG loc, SlayerSpecialMove sspm)
        {
            if (sspm == null)
                return;
            SpecialMoveBase effectInstance = GetSpecialMoveInstance(sspm.effect);
            if (effectInstance == null)
                return;
            if (effectInstance.GetLocationRG == loc)
            {
                effectInstance.DoSpecialMove(range, sspm.args);
            }
        }
    }

    public abstract class SpecialMoveBase
    {
        public const string RACE_HUMANLIKE = "Humanlike";
        public const string RACE_ANIMAL = "Animal";
        public const string RACE_MECHANOID = "Mechanoid";
        public const string RACE_INSECTOID = "Insectoid";

        public abstract SpecialMoveLocationML GetLocationML
        {
            get;
        }
        public abstract SpecialMoveLocationRG GetLocationRG
        {
            get;
        }
        public abstract void DoSpecialMove(SWB_Verb_MeleeAttack melee, List<string> args);
        public abstract void DoSpecialMove(SWB_Verb_RangedAttack range, List<string> args);
    }

    public class SPM_Delica43 : SpecialMoveBase
    {
        public override SpecialMoveLocationML GetLocationML
        {
            get
            {
                return SpecialMoveLocationML.BEFORE_END;
            }
        }
        public override SpecialMoveLocationRG GetLocationRG
        {
            get
            {
                return SpecialMoveLocationRG.NONE;
            }
        }
        public override void DoSpecialMove(SWB_Verb_MeleeAttack melee, List<string> args)
        {
            Pawn p = melee.CasterPawn;
            if (p == null)
                return;
            LocalTargetInfo target = melee.CurrentTarget;
            if (target == null)
                return;
            Pawn pt = target.Thing as Pawn;
            if (pt == null)
                return;

            bool ok = args == null || args.IsEmpty();
            if (!ok)
            {
                foreach (string defName in args)
                {
                    if (defName == pt.def.defName)
                    {
                        ok = true;
                        break;
                    }
                    if (pt.RaceProps == null)
                        break;
                    if (defName == RACE_HUMANLIKE && pt.RaceProps.Humanlike)
                    {
                        ok = true;
                        break;
                    }
                    if (defName == RACE_ANIMAL && pt.RaceProps.Animal)
                    {
                        ok = true;
                        break;
                    }
                    if (defName == RACE_MECHANOID && pt.RaceProps.IsMechanoid)
                    {
                        ok = true;
                        break;
                    }
                    if (defName == RACE_INSECTOID && pt.RaceProps.FleshType == FleshTypeDefOf.Insectoid)
                    {
                        ok = true;
                        break;
                    }
                }
            }

            if (!ok)
                return;

            p.needs.food.CurLevel = Math.Min(p.needs.food.MaxLevel, p.needs.food.CurLevel + 0.05f);

        }

        public override void DoSpecialMove(SWB_Verb_RangedAttack range, List<string> args)
        {
            // NONE
        }
    }


    public class SPM_ReloadOnEnemeyDown : SpecialMoveBase
    {

        public override SpecialMoveLocationML GetLocationML
        {
            get
            {
                return SpecialMoveLocationML.BEFORE_END;
            }
        }

        public override SpecialMoveLocationRG GetLocationRG
        {
            get
            {
                return SpecialMoveLocationRG.BEFORE_END;
            }
        }

        public override void DoSpecialMove(SWB_Verb_MeleeAttack melee, List<string> args)
        {
            Pawn p = melee.CasterPawn;
            if (p == null)
                return;
            string defaultEmoteText = "+Reload";
            if (!args.NullOrEmpty())
            {
                defaultEmoteText = args.First();
            }
            LocalTargetInfo target = melee.CurrentTarget;
            if (target == null)
                return;
            Pawn pt = target.Thing as Pawn;
            if (pt == null)
                return;

            if (pt.Destroyed || pt.Dead || pt.Downed)
            {
                melee.m_AttackCount = 0;
                Utility.ShowEmote(p, defaultEmoteText);
            }
        }

        public override void DoSpecialMove(SWB_Verb_RangedAttack range, List<string> args)
        {
            Pawn p = range.CasterPawn;
            if (p == null)
                return;
            string defaultEmoteText = "+Reload";
            if (!args.NullOrEmpty())
            {
                defaultEmoteText = args.First();
            }
            LocalTargetInfo target = range.CurrentTarget;
            if (target == null)
                return;
            Pawn pt = target.Thing as Pawn;
            if (pt == null)
                return;

            if (pt.Destroyed || pt.Dead || pt.Downed)
            {
                range.m_AttackCount = range.m_AddAttackNum;
                Utility.ShowEmote(p, defaultEmoteText);
            }
        }
    }

    public class SPM_ReloadOnEnemeyDownML : SPM_ReloadOnEnemeyDown
    {
        public override SpecialMoveLocationRG GetLocationRG
        {
            get
            {
                return SpecialMoveLocationRG.NONE;
            }
        }
    }
    public class SPM_ReloadOnEnemeyDownRG : SPM_ReloadOnEnemeyDown
    {
        public override SpecialMoveLocationML GetLocationML
        {
            get
            {
                return SpecialMoveLocationML.NONE;
            }
        }
    }

}
